/*
 * clock.c -- clock interrupt server.  In the ultimate version, this will
 * do the standard Z-80 interrupt things, but for now it will update the
 * screen.
 *
 * Notes:
 *	Z-80 interrupt generation is not handled very well yet; we just send
 * a Z-80 interrupt not caring if the rate is right or not (25 ms).
 *
 *	The clock interrupt handler can be vectored directly from the
 * clock interrupt or through the DOS hook INT 0x1C.  #define DIRECT_VECTOR
 * if you want it to be direct.  You want it to be direct, but here
 * are the evils we face:
 *
 * INT 0x1C hook
 *	The real-time clock won't get as screwed up (but it still will).  The
 * disk will stop spinning, too.  But the interrupt servicing is much
 * slower and interrupts are (apparently) disabled so keystrokes end up
 * getting lost if they occur during a screen update.  This is bad.
 *
 * Direct hook
 *	Real-time clock is totally screwed, but interrupts are serviced faster
 * and we don't drop keys.  The disk drive motor time-out also goes
 * bye-bye, so we have some kludges to not the simulator until a timeout
 * happens (ugh :-().
 *
 * What I really need to do is set up a direct vector and pass control
 * onto the DOS interrupt handler so it still gets 18 a second.  That
 * way, the time and the disk motor (and prob. some other stuff) is
 * happy but (hopefully) dropped keystrokes and overhead are not a
 * problem.
 */

#define DIRECT_VECTOR

#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "mtype.h"
#include "clock.h"
#include "z80.h"

int screen_refreshing = 0;
int refresh_overrun = 0;

long clock_tick;
/* to be fazed out, useless for other's purposes anyway */
float clock_freq = 19;
static int ips = 19;
static int sups = 19;
static int refresh_count;

static void interrupt clock_int();
static void interrupt (*oldclockvec)();

static int clock_installed = 0;

/*extern*/ unsigned char port_e0;

extern void screen_refresh();

int did_intr;

void clock_install(void)
{
	int	i;

#ifdef DIRECT_VECTOR
	oldclockvec = getvect(8);
#else
	oldclockvec = getvect(0x1c);
#endif

	clock_tick = 0;

	clock_div(ips2div((float)(sups = ips = 19)));
	refresh_count = 0;

#ifdef DIRECT_VECTOR
	setvect(8, clock_int);
#else
	setvect(0x1c, clock_int);
#endif

	clock_installed = 1;

	port_e0 = 255;
}

void clock_remove(void)
{
	if (clock_installed) {
#ifdef DIRECT_VECTOR
		setvect(8, oldclockvec);
#else
		setvect(0x1c, oldclockvec);
#endif
		clock_div(0);
	}
	clock_installed = 0;
}


void clock_div(int divisor)
{
	disable();
	outportb(0x43, 0x36);					/* select timer 0,lsb,msb,mode 3 */
	outportb(0x40, divisor & 255);			/* LSB of divisor */
	outportb(0x40, (divisor >> 8) & 255);	/* MSB of divisor */
	enable();
}

static void interrupt clock_int()
{
	register int i;
	register char ch;
	int dl;

	/* if the previous clock interrupt was ACKed, send off another one */
	/* we don't care if the interrupts are enabled, we just send the
	 * interrupt.
	 *
	 * If debugging mode is on, don't do interrupts and give the poor
	 * sod a chance.
	 *
	 * Interesting, interrupts screw up castle.cmd.  Have to see if they EI
	 * or not...
	 */
#ifdef CASTLE
	if ((port_e0 & 4) && !debug_on) {
		int_line = 1;
		port_e0 &= ~4;
		z80zap();
	}
#endif

	enable();

	clock_tick++;

	did_intr = 1;

#ifdef DIRECT_VECTOR
	outportb(0x20, 0x20);	/* acknowledge interrupt */
#endif

	if ((refresh_count += sups) < ips)
		return;

	refresh_count -= ips;

	if (screen_refreshing)
		refresh_overrun++;
	else {
		screen_refreshing++;
		screen_refresh();
		screen_refreshing--;
	}

	/*outportb(0x20, 0x20);*/	/* acknowledge interrupt */
}

#ifdef NOTNOW
static int int_count = 0;

/* maybe put this back later */
/*
 * This is an attempt at an interrupt routine which takes a move uniform
 * amount of time.  It does the "smart" update but every 6 interrupts or
 * so it does a "dumb" update.
 *
 * Probably won't bother to make an assembler version of this since
 * it is supposed to be slow!
 */
static void interrupt clock_intD()
{
	register int i;
	register int ch;

	clock_tick++;

	if (++int_count >= 6) {
		for (i = 0; i < 1024; i++) {
			screen_char(i, ch = mem_low[i + 15360]);
			screen[i] = ch;
		}
		int_count = 0;
	}
	else {
		for (i = 0; i < 1024; i++)
			if (screen[i] != (ch = mem_low[i + 15360])) {
				screen_char(i, ch);
				screen[i] = ch;
			}
	}

	outportb(0x20, 0x20);	/* acknowledge interrupt */
}
#endif