/*
 * key.c -- TRS-80 keyboard simulator (basically an interrupt handler
 * for the keyboard)
 *
 * BUG: if you map a key to a special function which does not have any
 * control functions associated with it, crash city.  This should
 * be fixed with a check....
 *
 * Seems that the 3800 matrix is duplicated at 3900 and 3a00 and 3b00.
 * I'll throw those in even though it will make key processing yet slower.
 * The machine language version will make things better and I'll make
 * all non-standard matrix simulations an option.
 *
 * The keyboard processing is too slow.  Assembler rewrite would help, but
 * I don't think it would solve the problem.  I will fix this by making a
 * queue of incoming keys.  With a little mutual exclusion, we guarantee
 * that there is at most 1 copy of the memory modifing loop running at
 * any moment.
 *
 * Based on the theory that key repeat is causing keys to stick on Peter's
 * machine, I'm putting in a table that indicates if we've already received
 * the make code for a key.  If so, do nothing.  When we get the break code,
 * indicate we will do something with the make code.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>
#include <dos.h>
#include "mtype.h"
#include "key.h"
#include "z80.h"
#include "trs80.h"

#define TK_ENTER '\r'
#define TK_CLEAR 'C'
#define TK_BREAK 'B'
#define TK_UP    'U'
#define TK_DOWN  'D'
#define TK_LEFT  'L'
#define TK_RIGHT 'R'
#define TK_LEFT_SHIFT 'X'
#define TK_RIGHT_SHIFT 'Y'
#define TK_NONE  'Z'

static char key_matrix[] = {
	'@', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
	'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
	'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
	'x', 'y', 'z', TK_NONE, TK_NONE, TK_NONE, TK_NONE, TK_NONE,
	'0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', ':', ';', ',', '-', '.', '/',
	TK_ENTER, TK_CLEAR, TK_BREAK, TK_UP, TK_DOWN, TK_LEFT, TK_RIGHT, ' ',
	TK_LEFT_SHIFT, TK_RIGHT_SHIFT
};

struct tkey {
	unsigned char rowbit;
	unsigned char colbit;
};

struct tkey tname_to_tkey[128];

void init_keyboard()
{
	char* km;
	int i, rowbit, colbit;

    for (i = 0; i < 128; i++)
		tname_to_tkey[i].rowbit = 0;

	rowbit = 1;
    colbit = 1;
	for (km = key_matrix; *km; km++) {
		if (*km != TK_NONE) {
			tname_to_tkey[*km].rowbit = rowbit;
			tname_to_tkey[*km].colbit = colbit;
		}
		colbit <<= 1;
		if (colbit == 256) {
			colbit = 1;
			rowbit <<= 1;
		}
    }
}

/* XXX -- it "knows" about the reference model, it shouldn't have to. */

static struct trs80* t80;
static mbyte* mem;

void press_tkey(struct tkey* key)
{
	if (key->rowbit) {
		mem[0x3800 + key->rowbit] |= key->colbit;
		if (t80->ref) {
			t80->ref->z80.mem[0x3800 + key->rowbit] |= key->colbit;
		}
    }
}

void release_tkey(struct tkey* key)
{
	if (key->rowbit) {
		mem[0x3800 + key->rowbit] &= ~key->colbit;
		if (t80->ref) {
			t80->ref->z80.mem[0x3800 + key->rowbit] &= ~key->colbit;
		}
    }
}

void release_all_keys(void)
{
	memset(mem + 0x3800, 0, 4 * 256);
	if (t80->ref) {
		memset(t80->ref->z80.mem + 0x3800, 0, 4 * 256);
	}
}

/* TRS-80 keyboard layout diagram:

        1 2 3 4 5 6 7 8 9 0 : - break      ! " # $ % & ' ( ) 0 * = break
       [ q w e r t y u i o p @ <- ->      [ Q W E R T Y U I O P @ <- ->
		] a s d f g h j k l ; enter clr   ] A S D F G H J K L + enter clr
         s z x c v b n m , . / s            s Z X C V B N M < > ? s
*/

/* some testing kinda stuff */
int key_was_running = 0;
int key_interrupts = 0;
int key_repeated = 0;
int	key_during_screen = 0;
extern int screen_refreshing;

/* kludgey stuff for the debugger */
volatile int scan_code = 0;

volatile int esc_hit = 0;

int term_key;

#define kb_data         (0x60)
#define kb_ctl          (0x61)
#define eoi             (0x20)
#define kbintnum        (9)

/* Interrupt vector and associated tables       */

static void interrupt key();
static void interrupt (*oldkeyvec)();

#define SHIFT(x)	((x) | 128)

/* format: PC key, physical mapping, WYSIWYG, WYSIWYG shifted */
unsigned char scanmaps[] = {
	/*   ` ~ */ 41, TK_CLEAR, TK_CLEAR, SHIFT(TK_CLEAR),
	/*   1 ! */ 2, '1', '1', SHIFT('1'),
	/*   2 @ */ 3, '2', '2', '@',
	/*   3 # */ 4, '3', '3', SHIFT('3'),
	/*   4 $ */ 5, '4', '4', SHIFT('4'),
	/*   5 % */ 6, '5', '5', SHIFT('5'),
	/*   6 ^ */ 7, '6', '6', 0,
	/*   7 & */ 8, '7', '7', SHIFT('6'),
	/*   8 * */ 9, '8', '8', SHIFT(':'),
	/*   9 ( */ 10, '9', '9', SHIFT('8'),
	/*   0 ) */ 11, '0', '0', SHIFT('9'),
	/*   - _ */ 12, ':', '-', 0,
	/*   = + */ 13, '-', SHIFT('-'), SHIFT(';'),
	/*   \ | */ 43, 0, 0, 0, /* XXX could physically go to break, I suppose */
	/*bkspace*/ 14, TK_LEFT, TK_LEFT, SHIFT(TK_LEFT),
	/*   [ { */ 26, TK_LEFT, TK_LEFT, SHIFT(TK_LEFT),
	/*   ] } */ 27, TK_RIGHT, TK_RIGHT, SHIFT(TK_RIGHT),
	/*   ; : */ 39, ';', ';', ':',
	/*   ' " */ 40, '@', SHIFT('7'), SHIFT('2'),
	/*   a   */ 30, 'a', 'a', SHIFT('a'),
	/*   b   */ 48, 'b', 'b', SHIFT('b'),
	/*   c   */ 46, 'c', 'c', SHIFT('c'),
	/*   d   */ 32, 'd', 'd', SHIFT('d'),
	/*   e   */ 18, 'e', 'e', SHIFT('e'),
	/*   f   */ 33, 'f', 'f', SHIFT('f'),
	/*   g   */ 34, 'g', 'g', SHIFT('g'),
	/*   h   */ 35, 'h', 'h', SHIFT('h'),
	/*   i   */ 23, 'i', 'i', SHIFT('i'),
	/*   j   */ 36, 'j', 'j', SHIFT('j'),
	/*   k   */ 37, 'k', 'k', SHIFT('k'),
	/*   l   */ 38, 'l', 'l', SHIFT('l'),
	/*   m   */ 50, 'm', 'm', SHIFT('m'),
	/*   n   */ 49, 'n', 'n', SHIFT('n'),
	/*   o   */ 24, 'o', 'o', SHIFT('o'),
	/*   p   */ 25, 'p', 'p', SHIFT('p'),
	/*   q   */ 16, 'q', 'q', SHIFT('q'),
	/*   r   */ 19, 'r', 'r', SHIFT('r'),
	/*   s   */ 31, 's', 's', SHIFT('s'),
	/*   t   */ 20, 't', 't', SHIFT('t'),
	/*   u   */ 22, 'u', 'u', SHIFT('u'),
	/*   v   */ 47, 'v', 'v', SHIFT('v'),
	/*   w   */ 17, 'w', 'w', SHIFT('w'),
	/*   x   */ 45, 'x', 'x', SHIFT('x'),
	/*   y   */ 21, 'y', 'y', SHIFT('y'),
	/*   z   */ 44, 'z', 'z', SHIFT('z'),
	/* space */ 57, ' ', ' ', SHIFT(' '),
	/* return*/ 28, TK_ENTER, TK_ENTER, SHIFT(TK_ENTER),
	/* lshift*/ 42, TK_LEFT_SHIFT, TK_LEFT_SHIFT, SHIFT(TK_LEFT_SHIFT),
	/* rshift*/ 54, TK_RIGHT_SHIFT, TK_RIGHT_SHIFT, SHIFT(TK_RIGHT_SHIFT),
	/*   tab */ 15, TK_UP, TK_RIGHT, SHIFT(TK_RIGHT),
	/* key 7 */ 71, '7', TK_CLEAR, SHIFT(TK_CLEAR),
	/* key 8 */ 72, '8', TK_UP, SHIFT(TK_UP),
	/* key 9 */ 73, '9', 0, 0,
	/* key 4 */ 75, '4', TK_LEFT, SHIFT(TK_LEFT),
	/* key 5 */ 76, '5', 0, 0,
	/* key 6 */ 77, '6', TK_RIGHT, SHIFT(TK_RIGHT),
	/* key 1 */ 79, '1', 0, 0,
	/* key 2 */ 80, '2', TK_DOWN, SHIFT(TK_DOWN),
	/* key 3 */ 81, '3', 0, 0,
	/*   , < */ 51, ',', ',', SHIFT(','),
	/*   . > */ 52, '.', '.', SHIFT('.'),
	/*   / ? */ 53, '/', '/', SHIFT('/'),
	/* ctrl  */ 29, TK_DOWN, TK_DOWN, SHIFT(TK_DOWN),
	/* break */ 70, TK_BREAK, TK_BREAK, SHIFT(TK_BREAK),
	/* delete*/ 83, TK_LEFT, TK_LEFT, SHIFT(TK_LEFT),
	/* insert*/ 82, '0', 0, 0,
	/* caps  */ 58, 0, SHIFT('0'), SHIFT('0'),
	/*numlock*/ 69, 0, 0, 0, /* XXX should really do different keypad stuff */
	/*     * */ 55, 0, 0, 0,
	/*     - */ 74, 0, 0, 0,
	/*     + */ 78, 0, 0, 0
};

unsigned char scan_to_name[128];
unsigned char shift_scan_to_name[128];
unsigned char phys_scan_to_name[128];

#define kb_data         (0x60)
#define kb_ctl          (0x61)
#define eoi             (0x20)
#define kbintnum        (9)

static void interrupt key();
static void interrupt (*oldkeyvec)();

static int kb_installed = 0;
static int shift;
static unsigned char press_code[128];
static int pcnt;
extern int physmap;

void kbinstall(struct trs80* t)
{
	int i;

	if (kb_installed)
		return;

	/* XXX */
	t80 = t;
	mem = t->z80.mem;

	init_keyboard();
	release_all_keys();

	for (i = 0; i < 128; i++)
		scan_to_name[i] = shift_scan_to_name[i] = phys_scan_to_name[i] = '\0';

	for (i = 0; i < sizeof(scanmaps); i += 4) {
		scan_to_name[scanmaps[i]] = scanmaps[i + 2];
		shift_scan_to_name[scanmaps[i]] = scanmaps[i + 3];
		phys_scan_to_name[scanmaps[i]] = scanmaps[i + 1];
	}

	shift = 0;
	pcnt = 0;
	for (i = 0; i < 128; i++)
		press_code[i] = 0;

	oldkeyvec = getvect(kbintnum);
	disable();
	setvect(kbintnum, key);
	kb_installed = 1;
	enable();
}

void kbremove(void)
{
	if (kb_installed)
		setvect(kbintnum, oldkeyvec);
	kb_installed = 0;
}

void ack_int(void);
void press_shift(void);
void release_shift(void);

static void interrupt key()
{
	int scan_code, trsshift;
	unsigned char* map, key;

	disable(); /* almost certainly unnecessary */

	scan_code = inportb(kb_data);

	if (scan_code == 1) {
		stop_z80();
		esc_hit = 1;
		ack_int();
		enable();
		return;
	}

	if (physmap) {
		key = phys_scan_to_name[scan_code & 0x7f];
		if (scan_code & 0x80) { /* break code */
			key &= 0x7f;
			release_tkey(tname_to_tkey + key);
		}
		else {
			press_tkey(tname_to_tkey + key);
		}
	}
	else {
		switch (scan_code & 127) {
		case 42: case 54: /* left shift, right shift */
			shift = !(scan_code & 128);
			if (pcnt == 0)
				if (shift) press_shift(); else release_shift();
			break;
		default:
			if (scan_code & 0x80) { /* break code */
				if (key = press_code[scan_code & 0x7f]) {
					pcnt--;
					trsshift = key & 0x80;
					key &= 0x7f;
					release_tkey(tname_to_tkey + key);
					press_code[scan_code & 0x7f] = 0;
					if (trsshift)
						release_shift();
				}
			}
			else { /* make code */
				if (!press_code[scan_code]) {
					map = shift ? shift_scan_to_name : scan_to_name;
					key = map[scan_code];
					press_code[scan_code] = key;
					pcnt++;
					trsshift = key & 0x80;
					key &= 0x7f;
					press_tkey(tname_to_tkey + key);
					if (trsshift)
						press_shift();
					else
						release_shift();
				}
			}
			if (pcnt == 0)
				if (shift) press_shift(); else release_shift();

			break;
		}
	}

	ack_int();
	enable();
}

void press_shift(void)
{
	press_tkey(tname_to_tkey + TK_LEFT_SHIFT);
	press_tkey(tname_to_tkey + TK_RIGHT_SHIFT);
}

void release_shift(void)
{
	release_tkey(tname_to_tkey + TK_LEFT_SHIFT);
	release_tkey(tname_to_tkey + TK_RIGHT_SHIFT);
}

void ack_int(void)
{
	register unsigned char a;

	a = inportb(kb_ctl);
	outportb(kb_ctl, a | 0x80);
	outportb(kb_ctl, a);
	disable();
	outportb(0x20, eoi);
}

void change_keyboard(void)
{
	/* could release all the keys */
}
