/*
 * load.c -- load ".hex", ".3bn" or ".cmd" files into
 * my simulated TRS 80.
 *
 * ".hex" files are from the z80 cross assembler called "zmac".
 * ".3bn" files are byte by byte copies of SYSTEM tape programs.
 * ".cmd" files are byte by byte copies of ".CMD" files from disk.
 * ".snp" snapshots of entire TRS-80 state.
 * ".bba" files are byte by byte copies of ".BAS" files from disk.
 * ".bas" files are ASCII (readable) dumps of BASIC programs.
 */

#include <stdio.h>
#include <alloc.h>
#include <conio.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "snap.h"
#include "load.h"
#include "rconfig.h"

int loaderr;
char loaderr_str[128];

/* buffer for use by load routines as long as they promise to be careful
 * about not hanging on to it too long.
 */
char apbuf[1024];

int abyte(char*, int*);
int aword(char*, int*);
int data_pack(char*, mbyte far*);

int load(struct trs80* t, char* fname)
{
	char* p;
	char* q = "";

	loaderr = 0;
	loaderr_str[0] = '\0';

	/* extract extension and call routine to handle that file type */
	/* :-(, a better routine would look for cookies! */
	/* XXX, I should at least look for snapshot cookie */
	for (p = fname; *p; p++)
		if (*p == '.')
			q = p;

	if (!strcmp(q, ".3bn"))
		return loadbin(t, fname);
	else if (!strcmp(q, ".cmd"))
		return loadcmd(t, fname);
	else if (!strcmp(q, ".snp"))
		return restoresnap(t, fname);
	else if (!strcmp(q, ".bas"))
		return loadbas(t, fname);
	else if (!strcmp(q, ".bba"))
		return loadbba(t, fname);
	else if (!strcmp(q, ".hex")) {
		return loadhex(t, fname);
	}
	else {
		loaderr = 1;
		strcpy(loaderr_str, "Don't know the file type.");
		return 0;
	}
}

int loadhex(struct trs80* t, char* fname)
{
	FILE* fp;
	int n;

	if (!(fp = path_fopen(fname, "r")))
		return 0;

	while (fgets(apbuf, 256, fp) != NULL) {
		if (!abyte(apbuf + 1, &n)) { fclose(fp); return 0; }
		if (n == 0) {
			fclose(fp);
			if (!aword(apbuf + 3, &n)) return 0;
			t->z80.reg.PC = n;
			return 1;
		}
		else {
			if (!data_pack(apbuf, t->z80.mem)) { fclose(fp); return 0; }
		}
	}
	fclose(fp);

	loaderr = 1;
	strcpy(loaderr_str, "No start address for .hex file.");
	return 0;
}

int data_pack(char* s, mbyte far* mem)
{
	unsigned int	st;
	int	n, i, b;

	if (!abyte(s + 1, &n)) return 0;
	if (!aword(s + 3, &st))	return 0;
	for (i = 9; n > 0; n--, i += 2) {
		if (!abyte(s + i, &b))
			return 0;
		mem[st++] = b;
	}
	return 1;
}

/* abyte -- take two byte hex string and return value */
int abyte(char* s, int* b)
{
	int d1, d2;

	if (!adig(s, &d1)) return 0;
	if (!adig(s + 1, &d2)) return 0;

	*b = (d1 << 4) | d2;

	return 1;
}

/* adig -- convert a hex digit to its proper value */
int adig(char* s, int* dig)
{
	char ch;

	ch = *s;
	if (ch >= '0' && ch <= '9')
		*dig = ch - '0';
	else if (ch >= 'A' && ch <= 'F')
		*dig = ch - 'A' + 10;
	else {
		loaderr = 1;
		strcpy(loaderr_str, "Bad hexadecimal digit.");
		return 0;
	}

	return 1;
}

/* aword -- convert a hex word in a four digit string */
int aword(char* s, int* w)
{
	int b1, b2;

	if (!abyte(s, &b1)) return 0;
	if (!abyte(s + 2, &b2)) return 0;

	*w = (b1 << 8) | b2;

	return 1;
}

void ueof(void)
{
	loaderr = 1;
	strcpy(loaderr_str, "Unexpected end of file.");
}

void loom(void)
{
	loaderr = 1;
	strcpy(loaderr_str, "Out of memory.");
}

/* Routines to handle opening a file given a list of directories to search */

struct path {
	char*	dir;
	struct path*	next;
};

static struct path*	path_list = NULL;
static struct path**	path_insert = &path_list;

void path_token(char* dir)
{
	struct path*	pth;
	char*			q;

	pth = (struct path*)malloc(sizeof(struct path));
	if (pth == NULL) {
		shutdown();
		fprintf(stderr, "out of memory");
		exit(1);
	}

	pth->dir = clonestr(dir);
	pth->next = NULL;

	*path_insert = pth;
	path_insert = &(pth->next);
}

FILE* path_fopen(char* fname, char* mode)
{
	FILE*	fp;
	struct path*	p;

	for (p = path_list; p != NULL; p = p->next) {
		strcpy(apbuf, p->dir);
		strcat(apbuf, "/");
		strcat(apbuf, fname);

		if ((fp = fopen(apbuf, mode)) != NULL)
			return fp;
	}

	if ((fp = fopen(fname, mode)) == NULL) {
		strcpy(loaderr_str, "Can't open file.");
		loaderr = LD_SYSERR;
		return 0;
	}

	return fp;
}
