/*
 * loadbas.c -- load a BASIC program stored in textual form.
 *
 * Not too much to do.  We have to tokenize any reserved words and
 * convert \ escapes in strings to characters.  The \ escapes are
 * just like C escapes.  There's a couple of oddities like the
 * ' abbreviation for REM and inserting colons before ELSE's.
 *
 * It would be very nice to have a version which allows you to save
 * a BASIC program in ASCII form (or even binary).
 *
 * As with loadbba, a major hack for now that has us load "basic.snp"
 * and then the BASIC program.  Will figure out how to do it elegantly
 * later!
 */

#include <stdio.h>
#include <alloc.h>
#include <ctype.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "snap.h"
#include "load.h"

static unsigned char tokenbuf[1024];

static int make_key_lookup(void);
static int tokenize(unsigned char*, unsigned char*);

int loadbas(struct trs80* t, char* fname)
{
	FILE*	fp;
	int		i;
	int		ch;
	int		len;
	unsigned int loadptr;

	if (!(fp = path_fopen(fname, "r")))
		return 0;

	if (!restoresnap(t, "basic.snp")) {
		fclose(fp);
		return 0;
	}

	if (!make_key_lookup()) {
		fclose(fp);
		return 0;
	}

	/* get pointer to start of BASIC program */
	loadptr = Peek16(0x40a4);

	while (fgets(apbuf, 1024, fp) != NULL) {
		if (len = tokenize(apbuf, tokenbuf)) {
			Poke16(loadptr, loadptr + 2 + len);
			loadptr += 2;
			for (i = 0; i < len; i++)
				Poke(loadptr++, tokenbuf[i]);
		}
	}
	Poke(loadptr++, 0);
	Poke(loadptr++, 0);

	/* now set up a few soft pointers */
	Poke16(0x40f9, loadptr);	/* end of BASIC program pointer */
	Poke16(0x40fb, loadptr);	/* start of array variables pointer */
	Poke16(0x40fd, loadptr);	/* start of free memory pointer */

	fclose(fp);

	return 1;
}

char* token[] = {
	"END",
	"FOR", "RESET", "SET", "CLS", "CMD", "RANDOM", "NEXT", "DATA", "INPUT",
	"DIM", "READ", "LET", "GOTO", "RUN", "IF", "RESTORE", "GOSUB", "RETURN",
	"REM", "STOP", "ELSE", "TRON", "TROFF", "DEFSTR", "DEFINT", "DEFSNG",
	"DEFDBL", "LINE", "EDIT", "ERROR", "RESUME", "OUT", "ON", "OPEN",
	"FIELD", "GET", "PUT", "CLOSE", "LOAD", "MERGE", "NAME", "KILL", "LSET",
	"RSET", "SAVE", "SYSTEM", "LPRINT", "DEF", "POKE", "PRINT", "CONT",
	"LIST", "LLIST", "DELETE", "AUTO", "CLEAR", "CLOAD", "CSAVE", "NEW",
	"TAB(", "TO", "FN", "USING", "VARPTR", "USR", "ERL", "ERR", "STRING$",
	"INSTR", "POINT", "TIME$", "MEM", "INKEY$", "THEN", "NOT", "STEP",
	"+", "-", "*", "/", "[", "AND", "OR", ">", "=", "<", "SGN", "INT",
	"ABS", "FRE", "INP", "POS", "SQR", "RND", "LOG", "EXP", "COS", "SIN",
	"TAN", "ATN", "PEEK", "CVI", "CVS", "CVD", "EOF", "LOC", "LOF", "MKI$",
	"MKS$", "MKD$", "CINT", "CSNG", "CDBL", "FIX", "LEN", "STR$", "VAL",
	"ASC", "CHR$", "LEFT$", "RIGHT$", "MID$", NULL };

struct tok {
	unsigned char*	fullname;
	int				len;
	int				token;
	struct tok*		next;
};

static int made_key_lookup = 0;
static struct tok* letter_to_tok[26];

/*
 * Create a table which will give us each reserved word that starts with
 * a certain letter.  This should make the tokenizing quite quick.
 */
static int make_key_lookup(void)
{
	int	i;
	int	n;
	struct tok* t;
	struct tok** tpp;
	char first;

	if (made_key_lookup)
		return 1;

	made_key_lookup = 1;

	for (n = 0; token[n] != NULL; n++)
		;

	if ((t = calloc(n, sizeof(struct tok))) == NULL) {
		loom();
		return 0;
	}

	for (i = 0; i < 26; i++)
		letter_to_tok[i] = NULL;

	for (i = 0; i < n; i++) {
		first = token[i][0];
		if (first < 'A' || first > 'Z')
			continue;
		first -= 'A';
		t[i].fullname = token[i];
		t[i].len = strlen(token[i]);
		t[i].token = i + 128;
		t[i].next = NULL;
		for (tpp = &(letter_to_tok[first]); *tpp != NULL; tpp = &((*tpp)->next))
			;
		*tpp = t + i;
	}

/*
	for (i = 0; i < 26; i++) {
		printf("%c: ", i + 'A');
		for (t = letter_to_tok[i]; t != NULL; t = t->next)
			printf("%s ", t->fullname);
		printf("\n");
	}
	exit(0);
*/
	return 1;
}

#define REM	(147)
#define DATA (136)
#define REMQUOT (251)
#define ELSE (149)

static int tokenize(unsigned char* line, unsigned char* tok)
{
	int	lnum;
	int	i;
	int	n;
	unsigned char*	tp;
	struct tok* t;

	tp = tok;

	while (isspace(*line))
		line++;

	lnum = atoi(line);

	if (!isdigit(*line))
		return 0;

	while (isdigit(*line))
		line++;

	while (isspace(*line))
		line++;

	if (!*line)
		return 0;

	*tp++ = lnum & 255;
	*tp++ = (lnum >> 8) & 255;

/*	printf("%d ", lnum); */

	while (*line) {
		switch (*line) {
		case '\n':
			line++;
			break;
		case '\'':
			*tp++ = ':';
			*tp++ = REM;
			*tp++ = REMQUOT;
			line++;
			while (*tp++ = *line++)
				;
			if (tp[-2] == '\n') {
				tp[-2] = '\0';
				tp--;
			}
			return tp - tok;
		case '"':
			*tp++ = *line++;
			while (*line && *line != '\n') {
				if (*line == '"') {
					*tp++ = *line++;
					break;
				}
				else if (*line == '\\') {
					line++;
					if (*line == 'n') {
						*tp++ = '\r';
						line++;
					}
					else if (*line >= '0' && *line <= '7') {
						n = 0;
						while (*line >= '0' && *line <= '7') {
							n *= 8;
							n += *line++ - '0';
						}
						*tp++ = n;
					}
					else
						*tp++ = '\\';
				}
				else
					*tp++ = *line++;
			}
			break;
		case '+':	*tp++ = 205; line++; break;
		case '-':	*tp++ = 206; line++; break;
		case '*':	*tp++ = 207; line++; break;
		case '/':	*tp++ = 208; line++; break;
		case '[':	*tp++ = 209; line++; break;
		case '>':	*tp++ = 212; line++; break;
		case '=':	*tp++ = 213; line++; break;
		case '<':	*tp++ = 214; line++; break;
		default:
			if (islower(*line))
				*line = toupper(*line);
			if (isupper(*line)) {
				*tp = *line;
				for (t = letter_to_tok[*line - 'A']; t != NULL; t = t->next) {
					if (!strnicmp(t->fullname, line, t->len)) {
						if (t->token == REM || t->token == DATA) {
							*tp++ = t->token;
							line += t->len;
							while (*tp++ = *line++)
								;
							if (tp[-2] == '\n') {
								tp[-2] = '\0';
								tp--;
							}
							return tp - tok;
						}
						else if (t->token == ELSE && tp[-1] != ':')
							*tp++ = ':';

						*tp = t->token;
						line += t->len - 1;
						break;
					}
				}
				tp++;
				line++;
			}
			else
				*tp++ = *line++;
		}
	}
	*tp++ = '\0';
	return tp - tok;
}
