/*
 * refz80.c -- a reference Z80 emulator implementation
 *
 * This is intended to aid debugging of other Z80 emulators; the
 * emphasis is on correctness, not speed.  Portability is useful,
 * but I sacrifice that a bit for some ease of implementation.
 *
 */

#include "mtype.h"
#include "z80.h"
#include "refz80.h"

/* You can #define BIG_ENDIAN or LITTLE_ENDIAN to make things faster,
 * but it is not necessary and is sanity checked anyways.  If you
 * don't know/don't care then don't bother.
 */
#define LITTLE_ENDIAN

/* values which hide little/big endian differences */
static int lowoff;
static int highoff;

#ifdef LITTLE_ENDIAN
#define low(w)	((mbyte*)&(w) + 0)
#define high(w)	((mbyte*)&(w) + 1)
#endif

#ifdef BIG_ENDIAN
#define low(w)	((mbyte*)&(w) + 1)
#define high(w)	((mbyte*)&(w) + 0)
#endif

#ifndef BIG_ENDIAN
#ifndef LITTLE_ENDIAN
#define low(w)	((mbyte*)&(w) + lowoff)
#define high(w)	((mbyte*)&(w) + highoff)
#endif
#endif

#define a	high(z->reg.AF)
#define f	low(z->reg.AF)
#define b	high(z->reg.BC)
#define c	low(z->reg.BC)
#define d	high(z->reg.DE)
#define e	low(z->reg.DE)
#define h	high(z->reg.HL)
#define l	low(z->reg.HL)

#define af	z->reg.AF
#define bc	z->reg.BC
#define de	z->reg.DE
#define hl	z->reg.HL
#define ix	z->reg.IX
#define iy	z->reg.IY
#define pc	z->reg.PC
#define sp	z->reg.SP
#define afp	z->reg.AFp
#define bcp	z->reg.BCp
#define dep	z->reg.DEp
#define hlp	z->reg.HLp

#define carry	(*f & 1)

static mbyte fetch_n(struct z80* z)
{
    return z->mem[pc++];
}

static mword fetch_nn(struct z80* z)
{
    mbyte lw;

    lw = z->mem[pc++];
    return (z->mem[pc++] << 8) | lw;
}

static void store16(struct z80* z, mword mem, mword val)
{
    z->mem[mem] = val & 0xff;
    z->mem[mem + 1] = (val >> 8) & 0xff;
}

static mword load16(struct z80* z, mword mem)
{
    mbyte lw;

    lw = z->mem[mem++];
    return (z->mem[mem] << 8) | lw;
}

static mbyte* ixy(struct z80* z, mword* xy)
{
	return z->mem + ((signed char)z->mem[pc++] + *xy);
}

/* utility routines for dealing with Z80 flag register */

static void assign_s(struct z80* z, int val)
{
    *f &= ~0x80;
    if (val) *f |= 0x80;
}

static void assign_z(struct z80* z, int val)
{
    *f &= ~0x40;
    if (val) *f |= 0x40;
}

static void assign_h(struct z80* z, int val)
{
    *f &= ~0x10;
    if (val) *f |= 0x10;
}

static void assign_pv(struct z80* z, int val)
{
    *f &= ~4;
    if (val) *f |= 4;
}

static void assign_n(struct z80* z, int val)
{
    *f &= ~2;
    if (val) *f |= 2;
}

static void assign_c(struct z80* z, int val)
{
    *f &= ~1;
    if (val) *f |= 1;
}

#define if_s	(*f & 0x80)
#define if_z	(*f & 0x40)
#define if_pv	(*f & 0x04)
#define if_c	(*f & 0x01)

static int parity(mbyte val)
{
    int cnt;

    cnt =
	!!(val & 0x80) +
	!!(val & 0x40) +
	!!(val & 0x20) +
	!!(val & 0x10) +
	!!(val & 0x08) +
	!!(val & 0x04) +
	!!(val & 0x02) +
	!!(val & 0x01);

    return !(cnt & 1);
}

/* routines which do most of the opcode implementation */

static void add(struct z80* z, mbyte* acc, int val, int cry, int docarry)
{
    int res;

    res = (int)*acc + val + cry;
    assign_s(z, res & 0x80);
    if (docarry) assign_c(z, res & 0x100);
    assign_h(z, ((*acc & 15) + (val & 15) + cry) & 16);
    assign_n(z, 0);

    assign_z(z, !(res & 0xff));

    res = (signed char)*acc + (signed char)val + cry;
    assign_pv(z, res > 127 || res < -128);

    *acc = (int)*acc + val + cry;
}

static void sub(struct z80* z, mbyte* acc, int val, int cry, int docarry)
{
    int res;

    res = (int)*acc - val - cry;
    assign_s(z, res & 0x80);
    if (docarry) assign_c(z, res & 0x100);
    assign_h(z, ((*acc & 15) - (val & 15) - cry) & 16);
    assign_n(z, 1);

    assign_z(z, !(res & 0xff));

    res = (signed char)*acc - (signed char)val - cry;
    assign_pv(z, res > 127 || res < -128);

    *acc = (int)*acc - val - cry;
}

static void cp(struct z80* z, int val)
{
    mbyte dummy = *a;
    sub(z, &dummy, val, 0, 1);
}

#define inc(r) add(z, r, 1, 0, 0)
#define dec(r) sub(z, r, 1, 0, 0)

static void and(struct z80* z, int val)
{
    *a &= val;
    assign_s(z, *a & 0x80);
    assign_z(z, !*a);
    assign_h(z, 1);
    assign_pv(z, parity(*a));
    assign_n(z, 0);
    assign_c(z, 0);
}

static void or(struct z80* z, int val)
{
    *a |= val;
    assign_s(z, *a & 0x80);
    assign_z(z, !*a);
    assign_h(z, 0);
    assign_pv(z, parity(*a));
    assign_n(z, 0);
    assign_c(z, 0);
}

static void xor(struct z80* z, int val)
{
    *a ^= val;
    assign_s(z, *a & 0x80);
    assign_z(z, !*a);
    assign_h(z, 0);
    assign_pv(z, parity(*a));
    assign_n(z, 0);
    assign_c(z, 0);
}

static void ex(mword* x, mword* y)
{
    mword temp;
    temp = *x;
    *x = *y;
    *y = temp;
}

static void ex_memsp(struct z80* z, mword* r)
{
    mword temp;

    temp = *r;
    *r = load16(z, sp);
    store16(z, sp, temp);
}

static void add16(struct z80* z, mword* acc, int val)
{
    int r;

    assign_n(z, 0);
    assign_h(z, ((*acc & 0xfff) + (val & 0xfff)) & 0x1000);
    /* Turbo C fucks this up; to it the hard way */
    /*res = (long)*acc + (long)val;*/
    r = (*low(*acc) + *low(val)) >> 8;
    assign_c(z, (*high(*acc) + *high(val) + r) & 0x100);
    *acc += val;
}

static void adc16(struct z80* z, mword* acc, int val)
{
    add(z, (mbyte*)acc + lowoff, val & 0xff, carry, 1);
    add(z, (mbyte*)acc + highoff, (val >> 8) & 0xff, carry, 1);
    assign_z(z, !*acc);
}

static void sbc16(struct z80* z, mword *acc, int val)
{
    sub(z, (mbyte*)acc + lowoff, val & 0xff, carry, 1);
    sub(z, (mbyte*)acc + highoff, (val >> 8) & 0xff, carry, 1);
    assign_z(z, !*acc);
}

static void daa(struct z80* z)
{
    int res = *a;

    if (*f & 2) {
	if ((res & 15) > 9 || (*f & 16))
	    res -= 6;
	if (*a > 0x99 || carry)
	    res -= 0x60;
	assign_h(z, (*f & 16) && (*a & 15) < 6);
	if (*a > 0x99)
	    assign_c(z, 1);
	*a = res;
    }
    else {
	if ((res & 15) > 9 || (*f & 16))
	    res += 6;
	if (res > 0x9f || carry)
	    res += 0x60;
	if (res > 255)
	    assign_c(z, 1);
	assign_h(z, (*a & 15) > 9);
	*a = res;
    }

    assign_s(z, *a & 0x80);
    assign_z(z, !*a);
    assign_pv(z, parity(*a));
}


static void rs_flags(struct z80* z, mbyte val)
{
    assign_z(z, !val);
    assign_s(z, val & 0x80);
    assign_h(z, 0);
    assign_pv(z, parity(val));
    assign_n(z, 0);
}

static void rlca(struct z80* z)
{
    assign_h(z, 0);
    assign_n(z, 0);
    assign_c(z, *a & 0x80);
    *a <<= 1;
    *a |= carry;
}

static void rla(struct z80* z)
{
    int prevc = carry;

    assign_h(z, 0);
    assign_n(z, 0);
    assign_c(z, *a & 0x80);
    *a <<= 1;
    *a |= prevc;
}

static void rrca(struct z80* z)
{
    assign_h(z, 0);
    assign_n(z, 0);
    assign_c(z, *a & 1);
    *a = (*a >> 1) & 0x7f;
    *a |= carry << 7;
}

static void rra(struct z80* z)
{
    int prevc = carry;

    assign_h(z, 0);
    assign_n(z, 0);
    assign_c(z, *a & 1);
    *a = (*a >> 1) & 0x7f;
    *a |= prevc << 7;
}

static void rlc(struct z80* z, mbyte* acc)
{
    assign_c(z, *acc & 0x80);
    *acc <<= 1;
    *acc |= carry;
    rs_flags(z, *acc);
}

static void rl(struct z80* z, mbyte* acc)
{
    int prevc = carry;

    assign_c(z, *acc & 0x80);
    *acc <<= 1;
    *acc |= prevc;
    rs_flags(z, *acc);
}

static void rrc(struct z80* z, mbyte* acc)
{
    assign_c(z, *acc & 1);
    /* note: *acc >>= 1 should work, but I avoid it because some C
     * compilers screw it up.
     */
    *acc = (*acc >> 1) & 0x7f;
    *acc |= carry << 7;
    rs_flags(z, *acc);
}

static void rr(struct z80* z, mbyte* acc)
{
    int prevc = carry;

    assign_c(z, *acc & 1);
    *acc = (*acc >> 1) & 0x7f;
    *acc |= prevc << 7;
    rs_flags(z, *acc);
}

static void sla(struct z80* z, mbyte* acc)
{
    assign_c(z, *acc & 0x80);
    *acc <<= 1;
    rs_flags(z, *acc);
}

static void sra(struct z80* z, mbyte* acc)
{
    assign_c(z, *acc & 1);
    /* again, I don't trust right shifts */
    *acc = (*acc >> 1) & 0x7f;
    *acc |= (*acc << 1) & 0x80;
    rs_flags(z, *acc);
}

static void srl(struct z80* z, mbyte* acc)
{
    assign_c(z, *acc & 1);
    *acc = (*acc >> 1) & 0x7f;
    rs_flags(z, *acc);
}

static void bit(struct z80* z, mbyte val, int bit)
{
    assign_h(z, 1);
    assign_n(z, 0);
    assign_z(z, !(val & (1 << bit)));
}

static void set(struct z80* z, mbyte* val, int bit)
{
    *val |= 1 << bit;
}

static void res(struct z80* z, mbyte* val, int bit)
{
    *val &= ~(1 << bit);
}

static void pop(struct z80* z, mword* val)
{
    *((mbyte*)val + lowoff) = z->mem[sp];
    sp++;
    *((mbyte*)val + highoff) = z->mem[sp];
    sp++;
}

static void push(struct z80* z, mword val)
{
    sp--;
    z->mem[sp] = (val >> 8) & 255;
    sp--;
    z->mem[sp] = val & 255;
}

static void jump(struct z80* z, int cond)
{
    mword newpc = fetch_nn(z);

    if (cond) pc = newpc;
}

static void branch(struct z80* z, int cond)
{
    mword newpc = (signed char)z->mem[pc++];

    newpc += pc;

    if (cond) pc = newpc;
}

static void call(struct z80* z, int cond)
{
    mword newpc = fetch_nn(z);

    if (cond) {
	push(z, pc);
	pc = newpc;
    }
}

/* IN + OUT are problematic and I'll have to think about them later */
static void in(struct z80* z, mbyte port, mbyte* acc, int doflags)
{
    *acc = 0xff;
    if (doflags) {
	assign_s(z, *acc & 0x80);
	assign_z(z, !*acc);
	assign_h(z, 0);
	assign_pv(z, parity(*acc));
	assign_n(z, 0);
    }
}

static void out(struct z80* z, mbyte port, mbyte val)
{
}

void initrefz80()
{
    mword x;

    x = 0x0100;
    lowoff = *(mbyte*)&x;
    x = 0x0001;
    highoff = *(mbyte*)&x;

#ifdef BIG_ENDIAN
    if (lowoff != 1) {
	printf("blew BIG_ENDIAN #ifdef\n");
	exit(1);
    }
#endif

#ifdef LITTLE_ENDIAN
    if (lowoff != 0) {
	printf("blew LITTLE_ENDIAN #ifdef\n");
	exit(1);
    }
#endif
}

static int main00(struct z80* z);
static int main01(struct z80* z);
static int main10(struct z80* z);
static int main11(struct z80* z);

int refz80(struct z80* z)
{
    /* big switches can be hard on C compilers so I break things up */
    switch (z->mem[pc] & 0xc0) {
    case 0x00: return main00(z);
    case 0x40: return main01(z);
    case 0x80: return main10(z);
    case 0xc0: return main11(z);
    }
}

static int main00(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x00: return 0;
    case 0x01: bc = fetch_nn(z); return 0;
    case 0x02: z->mem[bc] = *a; return MEM_CHG;
    case 0x03: bc++; return 0;
    case 0x04: inc(b); return 0;
    case 0x05: dec(b); return 0;
    case 0x06: *b = fetch_n(z); return 0;
    case 0x07: rlca(z); return 0;
    case 0x08: ex(&af, &afp); return 0;
    case 0x09: add16(z, &hl, bc); return 0;
    case 0x0a: *a = z->mem[bc]; return 0;
    case 0x0b: bc--; return 0;
    case 0x0c: inc(c); return 0;
    case 0x0d: dec(c); return 0;
    case 0x0e: *c = fetch_n(z); return 0;
    case 0x0f: rrca(z); return 0;
    case 0x10: (*b)--; branch(z, *b != 0); return 0;
    case 0x11: de = fetch_nn(z); return 0;
    case 0x12: z->mem[de] = *a; return MEM_CHG;
    case 0x13: de++; return 0;
    case 0x14: inc(d); return 0;
    case 0x15: dec(d); return 0;
    case 0x16: *d = fetch_n(z); return 0;
    case 0x17: rla(z); return 0;
    case 0x18: branch(z, 1); return 0;
    case 0x19: add16(z, &hl, de); return 0;
    case 0x1a: *a = z->mem[de]; return 0;
    case 0x1b: de--; return 0;
    case 0x1c: inc(e); return 0;
    case 0x1d: dec(e); return 0;
    case 0x1e: *e = fetch_n(z); return 0;
    case 0x1f: rra(z); return 0;
    case 0x20: branch(z, !if_z); return 0;
    case 0x21: hl = fetch_nn(z); return 0;
    case 0x22: store16(z, fetch_nn(z), hl); return MEM_CHG;
    case 0x23: hl++; return 0;
    case 0x24: inc(h); return 0;
    case 0x25: dec(h); return 0;
    case 0x26: *h = fetch_n(z); return 0;
    case 0x27: daa(z); return 0;
    case 0x28: branch(z, if_z); return 0;
    case 0x29: add16(z, &hl, hl); return 0;
    case 0x2a: hl = load16(z, fetch_nn(z)); return 0;
    case 0x2b: hl--; return 0;
    case 0x2c: inc(l); return 0;
    case 0x2d: dec(l); return 0;
    case 0x2e: *l = fetch_n(z); return 0;
    case 0x2f: *a = ~*a; assign_h(z, 1); assign_n(z, 1); return 0;
    case 0x30: branch(z, !if_c); return 0;
    case 0x31: sp = fetch_nn(z); return 0;
    case 0x32: z->mem[fetch_nn(z)] = *a; return MEM_CHG;
    case 0x33: sp++; return 0;
    case 0x34: inc(z->mem + hl); return MEM_CHG;
    case 0x35: dec(z->mem + hl); return MEM_CHG;
    case 0x36: z->mem[hl] = fetch_n(z); return MEM_CHG;
    case 0x37: assign_c(z, 1); assign_h(z, 0); assign_n(z, 0); return 0;
    case 0x38: branch(z, if_c); return 0;
    case 0x39: add16(z, &hl, sp); return 0;
    case 0x3a: *a = z->mem[fetch_nn(z)]; return 0;
    case 0x3b: sp--; return 0;
    case 0x3c: inc(a); return 0;
    case 0x3d: dec(a); return 0;
    case 0x3e: *a = fetch_n(z); return 0;
    case 0x3f: assign_h(z, carry); assign_c(z, !carry); assign_n(z, 0); return 0;
    }
}

static int main01(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x40: return 0;
    case 0x41: *b = *c; return 0;
    case 0x42: *b = *d; return 0;
    case 0x43: *b = *e; return 0;
    case 0x44: *b = *h; return 0;
    case 0x45: *b = *l; return 0;
    case 0x46: *b = z->mem[hl]; return 0;
    case 0x47: *b = *a; return 0;
    case 0x48: *c = *b; return 0;
    case 0x49: return 0;
    case 0x4a: *c = *d; return 0;
    case 0x4b: *c = *e; return 0;
    case 0x4c: *c = *h; return 0;
    case 0x4d: *c = *l; return 0;
    case 0x4e: *c = z->mem[hl]; return 0;
    case 0x4f: *c = *a; return 0;
    case 0x50: *d = *b; return 0;
    case 0x51: *d = *c; return 0;
    case 0x52: return 0;
    case 0x53: *d = *e; return 0;
    case 0x54: *d = *h; return 0;
    case 0x55: *d = *l; return 0;
    case 0x56: *d = z->mem[hl]; return 0;
    case 0x57: *d = *a; return 0;
    case 0x58: *e = *b; return 0;
    case 0x59: *e = *c; return 0;
    case 0x5a: *e = *d; return 0;
    case 0x5b: return 0;
    case 0x5c: *e = *h; return 0;
    case 0x5d: *e = *l; return 0;
    case 0x5e: *e = z->mem[hl]; return 0;
    case 0x5f: *e = *a; return 0;
    case 0x60: *h = *b; return 0;
    case 0x61: *h = *c; return 0;
    case 0x62: *h = *d; return 0;
    case 0x63: *h = *e; return 0;
    case 0x64: return 0;
    case 0x65: *h = *l; return 0;
    case 0x66: *h = z->mem[hl]; return 0;
    case 0x67: *h = *a; return 0;
    case 0x68: *l = *b; return 0;
    case 0x69: *l = *c; return 0;
    case 0x6a: *l = *d; return 0;
    case 0x6b: *l = *e; return 0;
    case 0x6c: *l = *h; return 0;
    case 0x6d: return 0;
    case 0x6e: *l = z->mem[hl]; return 0;
    case 0x6f: *l = *a; return 0;
    case 0x70: z->mem[hl] = *b; return MEM_CHG;
    case 0x71: z->mem[hl] = *c; return MEM_CHG;
    case 0x72: z->mem[hl] = *d; return MEM_CHG;
    case 0x73: z->mem[hl] = *e; return MEM_CHG;
    case 0x74: z->mem[hl] = *h; return MEM_CHG;
    case 0x75: z->mem[hl] = *l; return MEM_CHG;
    case 0x76: return HALT;
    case 0x77: z->mem[hl] = *a; return MEM_CHG;
    case 0x78: *a = *b; return 0;
    case 0x79: *a = *c; return 0;
    case 0x7a: *a = *d; return 0;
    case 0x7b: *a = *e; return 0;
    case 0x7c: *a = *h; return 0;
    case 0x7d: *a = *l; return 0;
    case 0x7e: *a = z->mem[hl]; return 0;
    case 0x7f: return 0;
    }
}

static int main10(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x80: add(z, a, *b, 0, 1); return 0;
    case 0x81: add(z, a, *c, 0, 1); return 0;
    case 0x82: add(z, a, *d, 0, 1); return 0;
    case 0x83: add(z, a, *e, 0, 1); return 0;
    case 0x84: add(z, a, *h, 0, 1); return 0;
    case 0x85: add(z, a, *l, 0, 1); return 0;
    case 0x86: add(z, a, z->mem[hl], 0, 1); return 0;
    case 0x87: add(z, a, *a, 0, 1); return 0;
    case 0x88: add(z, a, *b, carry, 1); return 0;
    case 0x89: add(z, a, *c, carry, 1); return 0;
    case 0x8a: add(z, a, *d, carry, 1); return 0;
    case 0x8b: add(z, a, *e, carry, 1); return 0;
    case 0x8c: add(z, a, *h, carry, 1); return 0;
    case 0x8d: add(z, a, *l, carry, 1); return 0;
    case 0x8e: add(z, a, z->mem[hl], carry, 1); return 0;
    case 0x8f: add(z, a, *a, carry, 1); return 0;
    case 0x90: sub(z, a, *b, 0, 1); return 0;
    case 0x91: sub(z, a, *c, 0, 1); return 0;
    case 0x92: sub(z, a, *d, 0, 1); return 0;
    case 0x93: sub(z, a, *e, 0, 1); return 0;
    case 0x94: sub(z, a, *h, 0, 1); return 0;
    case 0x95: sub(z, a, *l, 0, 1); return 0;
    case 0x96: sub(z, a, z->mem[hl], 0, 1); return 0;
    case 0x97: sub(z, a, *a, 0, 1); return 0;
    case 0x98: sub(z, a, *b, carry, 1); return 0;
    case 0x99: sub(z, a, *c, carry, 1); return 0;
    case 0x9a: sub(z, a, *d, carry, 1); return 0;
    case 0x9b: sub(z, a, *e, carry, 1); return 0;
    case 0x9c: sub(z, a, *h, carry, 1); return 0;
    case 0x9d: sub(z, a, *l, carry, 1); return 0;
    case 0x9e: sub(z, a, z->mem[hl], carry, 1); return 0;
    case 0x9f: sub(z, a, *a, carry, 1); return 0;
    case 0xa0: and(z, *b); return 0;
    case 0xa1: and(z, *c); return 0;
    case 0xa2: and(z, *d); return 0;
    case 0xa3: and(z, *e); return 0;
    case 0xa4: and(z, *h); return 0;
    case 0xa5: and(z, *l); return 0;
    case 0xa6: and(z, z->mem[hl]); return 0;
    case 0xa7: and(z, *a); return 0;
    case 0xa8: xor(z, *b); return 0;
    case 0xa9: xor(z, *c); return 0;
    case 0xaa: xor(z, *d); return 0;
    case 0xab: xor(z, *e); return 0;
    case 0xac: xor(z, *h); return 0;
    case 0xad: xor(z, *l); return 0;
    case 0xae: xor(z, z->mem[hl]); return 0;
    case 0xaf: xor(z, *a); return 0;
    case 0xb0: or(z, *b); return 0;
    case 0xb1: or(z, *c); return 0;
    case 0xb2: or(z, *d); return 0;
    case 0xb3: or(z, *e); return 0;
    case 0xb4: or(z, *h); return 0;
    case 0xb5: or(z, *l); return 0;
    case 0xb6: or(z, z->mem[hl]); return 0;
    case 0xb7: or(z, *a); return 0;
    case 0xb8: cp(z, *b); return 0;
    case 0xb9: cp(z, *c); return 0;
    case 0xba: cp(z, *d); return 0;
    case 0xbb: cp(z, *e); return 0;
    case 0xbc: cp(z, *h); return 0;
    case 0xbd: cp(z, *l); return 0;
    case 0xbe: cp(z, z->mem[hl]); return 0;
    case 0xbf: cp(z, *a); return 0;
    }
}

static int cb(struct z80*);
static int ed(struct z80*);
static int ixiy(struct z80*, mword*);

static int main11(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0xc0: if (!if_z) pop(z, &pc); return 0;
    case 0xc1: pop(z, &bc); return 0;
    case 0xc2: jump(z, !if_z); return 0;
    case 0xc3: jump(z, 1); return 0;
    case 0xc4: call(z, !if_z); return MEM_CHG;
    case 0xc5: push(z, bc); return MEM_CHG;
    case 0xc6: add(z, a, fetch_n(z), 0, 1); return 0;
    case 0xc7: push(z, pc); pc = 0; return MEM_CHG;
    case 0xc8: if (if_z) pop(z, &pc); return 0;
    case 0xc9: pop(z, &pc); return 0;
    case 0xca: jump(z, if_z); return 0;
    case 0xcb: return cb(z);
    case 0xcc: call(z, if_z); return MEM_CHG;
    case 0xcd: call(z, 1); return MEM_CHG;
    case 0xce: add(z, a, fetch_n(z), carry, 1); return 0;
    case 0xcf: push(z, pc); pc = 8; return MEM_CHG;
    case 0xd0: if (!if_c) pop(z, &pc); return 0;
    case 0xd1: pop(z, &de); return 0;
    case 0xd2: jump(z, !if_c); return 0;
    case 0xd3: out(z, fetch_n(z), *a); return IO_CHG;
    case 0xd4: call(z, !if_c); return MEM_CHG;
    case 0xd5: push(z, de); return MEM_CHG;
    case 0xd6: sub(z, a, fetch_n(z), 0, 1); return 0;
    case 0xd7: push(z, pc); pc = 0x10; return MEM_CHG;
    case 0xd8: if (if_c) pop(z, &pc); return 0;
    case 0xd9: ex(&bc, &bcp); ex(&de, &dep); ex(&hl, &hlp); return 0;
    case 0xda: jump(z, if_c); return 0;
    case 0xdb: in(z, fetch_n(z), a, 0); return IO_CHG;
    case 0xdc: call(z, if_c); return MEM_CHG;
    case 0xdd: return ixiy(z, &ix);
    case 0xde: sub(z, a, fetch_n(z), carry, 1); return 0;
    case 0xdf: push(z, pc); pc = 0x18; return MEM_CHG;
    case 0xe0: if (!if_pv) pop(z, &pc); return 0;
    case 0xe1: pop(z, &hl); return 0;
    case 0xe2: jump(z, !if_pv); return 0;
    case 0xe3: ex_memsp(z, &hl); return MEM_CHG;
    case 0xe4: call(z, !if_pv); return MEM_CHG;
    case 0xe5: push(z, hl); return MEM_CHG;
    case 0xe6: and(z, fetch_n(z)); return 0;
    case 0xe7: push(z, pc); pc = 0x20; return MEM_CHG;
    case 0xe8: if (if_pv) pop(z, &pc); return 0;
    case 0xe9: pc = hl; return 0;
    case 0xea: jump(z, if_pv); return 0;
    case 0xeb: ex(&de, &hl); return 0;
    case 0xec: call(z, if_pv); return MEM_CHG;
    case 0xed: return ed(z);
    case 0xee: xor(z, fetch_n(z)); return 0;
    case 0xef: push(z, pc); pc = 0x28; return MEM_CHG;
    case 0xf0: if (!if_s) pop(z, &pc); return 0;
    case 0xf1: pop(z, &af); return 0;
    case 0xf2: jump(z, !if_s); return 0;
    case 0xf3: z->reg.IFF1 = z->reg.IFF2 = 0; return 0;
    case 0xf4: call(z, !if_s); return 0;
    case 0xf5: push(z, af); return MEM_CHG;
    case 0xf6: or(z, fetch_n(z)); return 0;
    case 0xf7: push(z, pc); pc = 0x30; return MEM_CHG;
    case 0xf8: if (if_s) pop(z, &pc); return 0;
    case 0xf9: sp = hl; return 0;
    case 0xfa: jump(z, if_s); return 0;
    case 0xfb: z->reg.IFF1 = z->reg.IFF2 = 1; return 0;
    case 0xfc: call(z, if_s); return MEM_CHG;
    case 0xfd: return ixiy(z, &iy);
    case 0xfe: cp(z, fetch_n(z)); return 0;
    case 0xff: push(z, pc); pc = 0x38; return MEM_CHG;
    }
}

static int cb00(struct z80*);
static int cb01(struct z80*);
static int cb10(struct z80*);
static int cb11(struct z80*);

static int cb(struct z80* z)
{
    switch (z->mem[pc] & 0xc0) {
    case 0x00: return cb00(z);
    case 0x40: return cb01(z);
    case 0x80: return cb10(z);
    case 0xc0: return cb11(z);
    }
}

static int cb00(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x00: rlc(z, b); return 0;
    case 0x01: rlc(z, c); return 0;
    case 0x02: rlc(z, d); return 0;
    case 0x03: rlc(z, e); return 0;
    case 0x04: rlc(z, h); return 0;
    case 0x05: rlc(z, l); return 0;
    case 0x06: rlc(z, z->mem + hl); return MEM_CHG;
    case 0x07: rlc(z, a); return 0;
    case 0x08: rrc(z, b); return 0;
    case 0x09: rrc(z, c); return 0;
    case 0x0a: rrc(z, d); return 0;
    case 0x0b: rrc(z, e); return 0;
    case 0x0c: rrc(z, h); return 0;
    case 0x0d: rrc(z, l); return 0;
    case 0x0e: rrc(z, z->mem + hl); return MEM_CHG;
    case 0x0f: rrc(z, a); return 0;
    case 0x10: rl(z, b); return 0;
    case 0x11: rl(z, c); return 0;
    case 0x12: rl(z, d); return 0;
    case 0x13: rl(z, e); return 0;
    case 0x14: rl(z, h); return 0;
    case 0x15: rl(z, l); return 0;
    case 0x16: rl(z, z->mem + hl); return MEM_CHG;
    case 0x17: rl(z, a); return 0;
    case 0x18: rr(z, b); return 0;
    case 0x19: rr(z, c); return 0;
    case 0x1a: rr(z, d); return 0;
    case 0x1b: rr(z, e); return 0;
    case 0x1c: rr(z, h); return 0;
    case 0x1d: rr(z, l); return 0;
    case 0x1e: rr(z, z->mem + hl); return MEM_CHG;
    case 0x1f: rr(z, a); return 0;
    case 0x20: sla(z, b); return 0;
    case 0x21: sla(z, c); return 0;
    case 0x22: sla(z, d); return 0;
    case 0x23: sla(z, e); return 0;
    case 0x24: sla(z, h); return 0;
    case 0x25: sla(z, l); return 0;
    case 0x26: sla(z, z->mem + hl); return MEM_CHG;
    case 0x27: rlc(z, a); return 0;
    case 0x28: sra(z, b); return 0;
    case 0x29: sra(z, c); return 0;
    case 0x2a: sra(z, d); return 0;
    case 0x2b: sra(z, e); return 0;
    case 0x2c: sra(z, h); return 0;
    case 0x2d: sra(z, l); return 0;
    case 0x2e: sra(z, z->mem + hl); return MEM_CHG;
    case 0x2f: sra(z, a); return 0;
    case 0x38: srl(z, b); return 0;
    case 0x39: srl(z, c); return 0;
    case 0x3a: srl(z, d); return 0;
    case 0x3b: srl(z, e); return 0;
    case 0x3c: srl(z, h); return 0;
    case 0x3d: srl(z, l); return 0;
    case 0x3e: srl(z, z->mem + hl); return MEM_CHG;
    case 0x3f: srl(z, a); return 0;
    default: return ILL;
    }
}

static int cb01(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x40: bit(z, *b, 0); return 0;
    case 0x41: bit(z, *c, 0); return 0;
    case 0x42: bit(z, *d, 0); return 0;
    case 0x43: bit(z, *e, 0); return 0;
    case 0x44: bit(z, *h, 0); return 0;
    case 0x45: bit(z, *l, 0); return 0;
    case 0x46: bit(z, z->mem[hl], 0); return 0;
    case 0x47: bit(z, *a, 0); return 0;
    case 0x48: bit(z, *b, 1); return 0;
    case 0x49: bit(z, *c, 1); return 0;
    case 0x4a: bit(z, *d, 1); return 0;
    case 0x4b: bit(z, *e, 1); return 0;
    case 0x4c: bit(z, *h, 1); return 0;
    case 0x4d: bit(z, *l, 1); return 0;
    case 0x4e: bit(z, z->mem[hl], 1); return 0;
    case 0x4f: bit(z, *a, 1); return 0;
    case 0x50: bit(z, *b, 2); return 0;
    case 0x51: bit(z, *c, 2); return 0;
    case 0x52: bit(z, *d, 2); return 0;
    case 0x53: bit(z, *e, 2); return 0;
    case 0x54: bit(z, *h, 2); return 0;
    case 0x55: bit(z, *l, 2); return 0;
    case 0x56: bit(z, z->mem[hl], 2); return 0;
    case 0x57: bit(z, *a, 2); return 0;
    case 0x58: bit(z, *b, 3); return 0;
    case 0x59: bit(z, *c, 3); return 0;
    case 0x5a: bit(z, *d, 3); return 0;
    case 0x5b: bit(z, *e, 3); return 0;
    case 0x5c: bit(z, *h, 3); return 0;
    case 0x5d: bit(z, *l, 3); return 0;
    case 0x5e: bit(z, z->mem[hl], 3); return 0;
    case 0x5f: bit(z, *a, 3); return 0;
    case 0x60: bit(z, *b, 4); return 0;
    case 0x61: bit(z, *c, 4); return 0;
    case 0x62: bit(z, *d, 4); return 0;
    case 0x63: bit(z, *e, 4); return 0;
    case 0x64: bit(z, *h, 4); return 0;
    case 0x65: bit(z, *l, 4); return 0;
    case 0x66: bit(z, z->mem[hl], 4); return 0;
    case 0x67: bit(z, *a, 4); return 0;
    case 0x68: bit(z, *b, 5); return 0;
    case 0x69: bit(z, *c, 5); return 0;
    case 0x6a: bit(z, *d, 5); return 0;
    case 0x6b: bit(z, *e, 5); return 0;
    case 0x6c: bit(z, *h, 5); return 0;
    case 0x6d: bit(z, *l, 5); return 0;
    case 0x6e: bit(z, z->mem[hl], 5); return 0;
    case 0x6f: bit(z, *a, 5); return 0;
    case 0x70: bit(z, *b, 6); return 0;
    case 0x71: bit(z, *c, 6); return 0;
    case 0x72: bit(z, *d, 6); return 0;
    case 0x73: bit(z, *e, 6); return 0;
    case 0x74: bit(z, *h, 6); return 0;
    case 0x75: bit(z, *l, 6); return 0;
    case 0x76: bit(z, z->mem[hl], 6); return 0;
    case 0x77: bit(z, *a, 6); return 0;
    case 0x78: bit(z, *b, 7); return 0;
    case 0x79: bit(z, *c, 7); return 0;
    case 0x7a: bit(z, *d, 7); return 0;
    case 0x7b: bit(z, *e, 7); return 0;
    case 0x7c: bit(z, *h, 7); return 0;
    case 0x7d: bit(z, *l, 7); return 0;
    case 0x7e: bit(z, z->mem[hl], 7); return 0;
    case 0x7f: bit(z, *a, 7); return 0;
    }
}

static int cb10(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0x80: res(z, b, 0); return 0;
    case 0x81: res(z, c, 0); return 0;
    case 0x82: res(z, d, 0); return 0;
    case 0x83: res(z, e, 0); return 0;
    case 0x84: res(z, h, 0); return 0;
    case 0x85: res(z, l, 0); return 0;
    case 0x86: res(z, z->mem + hl, 0); return 0;
    case 0x87: res(z, a, 0); return 0;
    case 0x88: res(z, b, 1); return 0;
    case 0x89: res(z, c, 1); return 0;
    case 0x8a: res(z, d, 1); return 0;
    case 0x8b: res(z, e, 1); return 0;
    case 0x8c: res(z, h, 1); return 0;
    case 0x8d: res(z, l, 1); return 0;
    case 0x8e: res(z, z->mem + hl, 1); return 0;
    case 0x8f: res(z, a, 1); return 0;
    case 0x90: res(z, b, 2); return 0;
    case 0x91: res(z, c, 2); return 0;
    case 0x92: res(z, d, 2); return 0;
    case 0x93: res(z, e, 2); return 0;
    case 0x94: res(z, h, 2); return 0;
    case 0x95: res(z, l, 2); return 0;
    case 0x96: res(z, z->mem + hl, 2); return 0;
    case 0x97: res(z, a, 2); return 0;
    case 0x98: res(z, b, 3); return 0;
    case 0x99: res(z, c, 3); return 0;
    case 0x9a: res(z, d, 3); return 0;
    case 0x9b: res(z, e, 3); return 0;
    case 0x9c: res(z, h, 3); return 0;
    case 0x9d: res(z, l, 3); return 0;
    case 0x9e: res(z, z->mem + hl, 3); return 0;
    case 0x9f: res(z, a, 3); return 0;
    case 0xa0: res(z, b, 4); return 0;
    case 0xa1: res(z, c, 4); return 0;
    case 0xa2: res(z, d, 4); return 0;
    case 0xa3: res(z, e, 4); return 0;
    case 0xa4: res(z, h, 4); return 0;
    case 0xa5: res(z, l, 4); return 0;
    case 0xa6: res(z, z->mem + hl, 4); return 0;
    case 0xa7: res(z, a, 4); return 0;
    case 0xa8: res(z, b, 5); return 0;
    case 0xa9: res(z, c, 5); return 0;
    case 0xaa: res(z, d, 5); return 0;
    case 0xab: res(z, e, 5); return 0;
    case 0xac: res(z, h, 5); return 0;
    case 0xad: res(z, l, 5); return 0;
    case 0xae: res(z, z->mem + hl, 5); return 0;
    case 0xaf: res(z, a, 5); return 0;
    case 0xb0: res(z, b, 6); return 0;
    case 0xb1: res(z, c, 6); return 0;
    case 0xb2: res(z, d, 6); return 0;
    case 0xb3: res(z, e, 6); return 0;
    case 0xb4: res(z, h, 6); return 0;
    case 0xb5: res(z, l, 6); return 0;
    case 0xb6: res(z, z->mem + hl, 6); return 0;
    case 0xb7: res(z, a, 6); return 0;
    case 0xb8: res(z, b, 7); return 0;
    case 0xb9: res(z, c, 7); return 0;
    case 0xba: res(z, d, 7); return 0;
    case 0xbb: res(z, e, 7); return 0;
    case 0xbc: res(z, h, 7); return 0;
    case 0xbd: res(z, l, 7); return 0;
    case 0xbe: res(z, z->mem + hl, 7); return 0;
    case 0xbf: res(z, a, 7); return 0;
    }
}

static int cb11(struct z80* z)
{
    switch (z->mem[pc++]) {
    case 0xc0: set(z, b, 0); return 0;
    case 0xc1: set(z, c, 0); return 0;
    case 0xc2: set(z, d, 0); return 0;
    case 0xc3: set(z, e, 0); return 0;
    case 0xc4: set(z, h, 0); return 0;
    case 0xc5: set(z, l, 0); return 0;
    case 0xc6: set(z, z->mem + hl, 0); return 0;
    case 0xc7: set(z, a, 0); return 0;
    case 0xc8: set(z, b, 1); return 0;
    case 0xc9: set(z, c, 1); return 0;
    case 0xca: set(z, d, 1); return 0;
    case 0xcb: set(z, e, 1); return 0;
    case 0xcc: set(z, h, 1); return 0;
    case 0xcd: set(z, l, 1); return 0;
    case 0xce: set(z, z->mem + hl, 1); return 0;
    case 0xcf: set(z, a, 1); return 0;
    case 0xd0: set(z, b, 2); return 0;
    case 0xd1: set(z, c, 2); return 0;
    case 0xd2: set(z, d, 2); return 0;
    case 0xd3: set(z, e, 2); return 0;
    case 0xd4: set(z, h, 2); return 0;
    case 0xd5: set(z, l, 2); return 0;
    case 0xd6: set(z, z->mem + hl, 2); return 0;
    case 0xd7: set(z, a, 2); return 0;
    case 0xd8: set(z, b, 3); return 0;
    case 0xd9: set(z, c, 3); return 0;
    case 0xda: set(z, d, 3); return 0;
    case 0xdb: set(z, e, 3); return 0;
    case 0xdc: set(z, h, 3); return 0;
    case 0xdd: set(z, l, 3); return 0;
    case 0xde: set(z, z->mem + hl, 3); return 0;
    case 0xdf: set(z, a, 3); return 0;
    case 0xe0: set(z, b, 4); return 0;
    case 0xe1: set(z, c, 4); return 0;
    case 0xe2: set(z, d, 4); return 0;
    case 0xe3: set(z, e, 4); return 0;
    case 0xe4: set(z, h, 4); return 0;
    case 0xe5: set(z, l, 4); return 0;
    case 0xe6: set(z, z->mem + hl, 4); return 0;
    case 0xe7: set(z, a, 4); return 0;
    case 0xe8: set(z, b, 5); return 0;
    case 0xe9: set(z, c, 5); return 0;
    case 0xea: set(z, d, 5); return 0;
    case 0xeb: set(z, e, 5); return 0;
    case 0xec: set(z, h, 5); return 0;
    case 0xed: set(z, l, 5); return 0;
    case 0xee: set(z, z->mem + hl, 5); return 0;
    case 0xef: set(z, a, 5); return 0;
    case 0xf0: set(z, b, 6); return 0;
    case 0xf1: set(z, c, 6); return 0;
    case 0xf2: set(z, d, 6); return 0;
    case 0xf3: set(z, e, 6); return 0;
    case 0xf4: set(z, h, 6); return 0;
    case 0xf5: set(z, l, 6); return 0;
    case 0xf6: set(z, z->mem + hl, 6); return 0;
    case 0xf7: set(z, a, 6); return 0;
    case 0xf8: set(z, b, 7); return 0;
    case 0xf9: set(z, c, 7); return 0;
    case 0xfa: set(z, d, 7); return 0;
    case 0xfb: set(z, e, 7); return 0;
    case 0xfc: set(z, h, 7); return 0;
    case 0xfd: set(z, l, 7); return 0;
    case 0xfe: set(z, z->mem + hl, 7); return 0;
    case 0xff: set(z, a, 7); return 0;
    }
}

static int ed(struct z80* z)
{
    mbyte t;

    switch (z->mem[pc++]) {
    case 0x40: in(z, *c, b, 1); return IO_CHG;
    case 0x41: out(z, *c, *b); return IO_CHG;
    case 0x42: sbc16(z, &hl, bc); return 0;
    case 0x43: store16(z, fetch_nn(z), bc); return MEM_CHG;
    case 0x44: t = *a; *a = 0; sub(z, a, t, 0, 1); return 0;
    case 0x45: z->reg.IFF1 = z->reg.IFF2; pop(z, &pc); return 0;
    case 0x46: z->reg.IM = 0; return 0;
    case 0x47: z->reg.I = *a; return 0;
    case 0x48: in(z, *c, c, 1); return IO_CHG;
    case 0x49: out(z, *c, *c); return IO_CHG;
    case 0x4a: adc16(z, &hl, bc); return 0;
    case 0x4b: bc = load16(z, fetch_nn(z)); return 0;
    case 0x4d: pop(z, &pc); return 0; /* int acknowledge? */
    case 0x4f: z->reg.R = *a; return 0;
    case 0x50: in(z, *c, d, 1); return IO_CHG;
    case 0x51: out(z, *c, *d); return IO_CHG;
    case 0x52: sbc16(z, &hl, de); return 0;
    case 0x53: store16(z, fetch_nn(z), de); return MEM_CHG;
    case 0x56: z->reg.IM = 1; return 0;
    case 0x57:
	*a = z->reg.I;
	assign_h(z, 0);
	assign_n(z, 0);
	assign_s(z, z->reg.I & 0x80);
	assign_z(z, !z->reg.I);
	assign_pv(z, z->reg.IFF2);
	return 0;
    case 0x58: in(z, *c, e, 1); return IO_CHG;
    case 0x59: out(z, *c, *e); return IO_CHG;
    case 0x5a: adc16(z, &hl, de); return 0;
    case 0x5b: de = load16(z, fetch_nn(z)); return 0;
    case 0x5e: z->reg.IM = 2; return 0;
    case 0x5f:
	*a = z->reg.R;
	assign_h(z, 0);
	assign_n(z, 0);
	assign_s(z, z->reg.R & 0x80);
	assign_z(z, !z->reg.R);
	assign_pv(z, z->reg.IFF2);
	return 0;
    case 0x60: in(z, *c, h, 1); return IO_CHG;
    case 0x61: out(z, *c, *h); return IO_CHG;
    case 0x62: sbc16(z, &hl, hl); return 0;
    case 0x67:
	t = z->mem[hl] & 15;
	z->mem[hl] = ((z->mem[hl] >> 4) & 15) | (*a << 4);
	*a = (*a & 0xf0) | t;
	assign_s(z, *a & 0x80);
	assign_z(z, !*a);
	assign_h(z, 0);
	assign_pv(z, parity(*a));
	assign_n(z, 0);
	return MEM_CHG;
    case 0x68: in(z, *c, l, 1); return IO_CHG;
    case 0x69: out(z, *c, *l); return IO_CHG;
    case 0x6a: adc16(z, &hl, hl); return 0;
    case 0x6f:
	t = (z->mem[hl] >> 4) & 15;
	z->mem[hl] = (z->mem[hl] << 4) | (*a & 15);
	*a = (*a & 0xf0) | t;
	assign_s(z, *a & 0x80);
	assign_z(z, !*a);
	assign_h(z, 0);
	assign_pv(z, parity(*a));
	assign_n(z, 0);
	return MEM_CHG;
    case 0x72: sbc16(z, &hl, sp); return 0;
    case 0x73: store16(z, fetch_nn(z), sp); return MEM_CHG;
    case 0x78: in(z, *c, a, 1); return IO_CHG;
    case 0x79: out(z, *c, *a); return IO_CHG;
    case 0x7a: adc16(z, &hl, sp); return 0;
    case 0x7b: sp = load16(z, fetch_nn(z)); return 0;
    case 0xa0:
	z->mem[de++] = z->mem[hl++];
	bc--;
	assign_h(z, 0);
	assign_n(z, 0);
	assign_pv(z, bc);
	return MEM_CHG;
    case 0xa1:
	t = carry;
	cp(z, z->mem[hl++]);
	bc--;
	assign_pv(z, bc);
	assign_c(z, t);
	return 0;
    case 0xa2:
	in(z, *c, z->mem + hl++, 0);
	(*b)--;
	assign_z(z, !b);
	assign_n(z, 1);
	return IO_CHG | MEM_CHG;
    case 0xa3:
	out(z, *c, z->mem[hl++]);
	(*b)--;
	assign_z(z, !b);
	assign_n(z, 1);
	return IO_CHG | MEM_CHG;
    case 0xa8:
	z->mem[de--] = z->mem[hl--];
	bc--;
	assign_h(z, 0);
	assign_n(z, 0);
	assign_pv(z, bc);
	return MEM_CHG;
    case 0xa9:
	t = carry;
	cp(z, z->mem[hl--]);
	bc--;
	assign_pv(z, bc);
	assign_c(z, t);
	return 0;
    case 0xaa:
	in(z, *c, z->mem + hl--, 0);
	(*b)--;
	assign_z(z, !b);
	assign_n(z, 1);
	return IO_CHG | MEM_CHG;
    case 0xab:
	out(z, *c, z->mem[hl--]);
	(*b)--;
	assign_z(z, !b);
	assign_n(z, 1);
	return IO_CHG | MEM_CHG;
    case 0xb0:
	do {
	    z->mem[de++] = z->mem[hl++];
	    bc--;
	    assign_h(z, 0);
	    assign_n(z, 0);
	    assign_pv(z, bc);
	} while (bc);
	return MEM_CHG;
    case 0xb1:
	t = carry;
	do {
	    cp(z, z->mem[hl++]);
	    bc--;
	    assign_pv(z, bc);
	    assign_c(z, t);
	} while (bc && !if_z);
	return 0;
    case 0xb2:
	do {
	    in(z, *c, z->mem + hl++, 0);
	    (*b)--;
	    assign_z(z, !b);
	    assign_n(z, 1);
	} while (*b);
	return IO_CHG | MEM_CHG;
    case 0xb3:
	do {
	    out(z, *c, z->mem[hl++]);
	    (*b)--;
	    assign_z(z, !b);
	    assign_n(z, 1);
	} while (*b);
	return IO_CHG | MEM_CHG;
    case 0xb8:
	do {
	    z->mem[de--] = z->mem[hl--];
	    bc--;
	    assign_h(z, 0);
	    assign_n(z, 0);
	    assign_pv(z, bc);
	} while (bc);
	return MEM_CHG;
    case 0xb9:
	t = carry;
	do {
	    cp(z, z->mem[hl--]);
	    bc--;
	    assign_pv(z, bc);
	    assign_c(z, t);
	} while (bc && !if_z);
	return 0;
    case 0xba:
	do {
	    in(z, *c, z->mem + hl--, 0);
	    (*b)--;
	    assign_z(z, !b);
	    assign_n(z, 1);
	} while (*b);
	return IO_CHG | MEM_CHG;
    case 0xbb:
	do {
	    out(z, *c, z->mem[hl]);
	    (*b)--;
	    assign_z(z, !b);
	    assign_n(z, 1);
	} while (*b);
	return IO_CHG | MEM_CHG;
    default: return ILL;
    }
}

static int ixiy(struct z80* z, mword* xy)
{
    mbyte* i;

    switch (z->mem[pc++]) {
    case 0x09: add16(z, xy, bc); return 0;
    case 0x19: add16(z, xy, de); return 0;
    case 0x21: *xy = fetch_nn(z); return 0;
    case 0x22: store16(z, fetch_nn(z), *xy); return MEM_CHG;
    case 0x23: (*xy)++; return 0;
    case 0x29: add16(z, xy, *xy); return 0;
    case 0x2a: *xy = load16(z, fetch_nn(z)); return 0;
    case 0x2b: (*xy)--; return 0;
    case 0x34: inc(ixy(z, xy)); return MEM_CHG;
    case 0x35: dec(ixy(z, xy)); return MEM_CHG;
    case 0x36: i = ixy(z, xy); *i = fetch_n(z); return MEM_CHG;
    case 0x39: add16(z, xy, sp); return 0;
    case 0x46: *b = *ixy(z, xy); return 0;
    case 0x4e: *c = *ixy(z, xy); return 0;
    case 0x56: *d = *ixy(z, xy); return 0;
    case 0x5e: *e = *ixy(z, xy); return 0;
    case 0x66: *h = *ixy(z, xy); return 0;
    case 0x6e: *l = *ixy(z, xy); return 0;
    case 0x70: *ixy(z, xy) = *b; return MEM_CHG;
    case 0x71: *ixy(z, xy) = *c; return MEM_CHG;
    case 0x72: *ixy(z, xy) = *d; return MEM_CHG;
    case 0x73: *ixy(z, xy) = *e; return MEM_CHG;
    case 0x74: *ixy(z, xy) = *h; return MEM_CHG;
    case 0x75: *ixy(z, xy) = *l; return MEM_CHG;
    case 0x77: *ixy(z, xy) = *a; return MEM_CHG;
    case 0x7e: *a = *ixy(z, xy); return 0;
    case 0x86: add(z, a, *ixy(z, xy), 0, 1); return 0;
    case 0x8e: add(z, a, *ixy(z, xy), carry, 1); return 0;
    case 0x96: sub(z, a, *ixy(z, xy), 0, 1); return 0;
    case 0x9e: sub(z, a, *ixy(z, xy), carry, 1); return 0;
    case 0xa6: and(z, *ixy(z, xy)); return 0;
    case 0xae: xor(z, *ixy(z, xy)); return 0;
    case 0xb6: or(z, *ixy(z, xy)); return 0;
    case 0xbe: cp(z, *ixy(z, xy)); return 0;
    case 0xe1: pop(z, xy); return 0;
    case 0xe3: ex_memsp(z, xy); return MEM_CHG;
    case 0xe5: push(z, *xy); return MEM_CHG;
    case 0xe9: pc = *xy; return 0;
    case 0xf9: sp = *xy; return 0;
    default: return ILL;
    case 0xcb:
	i = ixy(z, xy);
	switch (z->mem[pc++]) {
	case 0x06: rlc(z, i); return MEM_CHG;
	case 0x0e: rrc(z, i); return MEM_CHG;
	case 0x16: rl(z, i); return MEM_CHG;
	case 0x1e: rr(z, i); return MEM_CHG;
	case 0x26: sla(z, i); return MEM_CHG;
	case 0x2e: sra(z, i); return MEM_CHG;
	case 0x3e: srl(z, i); return MEM_CHG;
	case 0x46: bit(z, *i, 0); return 0;
	case 0x4e: bit(z, *i, 1); return 0;
	case 0x56: bit(z, *i, 2); return 0;
	case 0x5e: bit(z, *i, 3); return 0;
	case 0x66: bit(z, *i, 4); return 0;
	case 0x6e: bit(z, *i, 5); return 0;
	case 0x76: bit(z, *i, 6); return 0;
	case 0x7e: bit(z, *i, 7); return 0;
	case 0x86: res(z, i, 0); return MEM_CHG;
	case 0x8e: res(z, i, 1); return MEM_CHG;
	case 0x96: res(z, i, 2); return MEM_CHG;
	case 0x9e: res(z, i, 3); return MEM_CHG;
	case 0xa6: res(z, i, 4); return MEM_CHG;
	case 0xae: res(z, i, 5); return MEM_CHG;
	case 0xb6: res(z, i, 6); return MEM_CHG;
	case 0xbe: res(z, i, 7); return MEM_CHG;
	case 0xc6: set(z, i, 0); return MEM_CHG;
	case 0xce: set(z, i, 1); return MEM_CHG;
	case 0xd6: set(z, i, 2); return MEM_CHG;
	case 0xde: set(z, i, 3); return MEM_CHG;
	case 0xe6: set(z, i, 4); return MEM_CHG;
	case 0xee: set(z, i, 5); return MEM_CHG;
	case 0xf6: set(z, i, 6); return MEM_CHG;
	case 0xfe: set(z, i, 7); return MEM_CHG;
	default: return ILL;
	}
    }
}
