/*
 * snap.c -- produce a "snapshot" of the simulated TRS-80's state.
 *
 * Uses a fancy tag-based format.  It can also read my emulator's
 * old format with was just the registers followed by all the memory.
 * That format is detected by file size alone and can be decyphered
 * from the code below.  See "snap.h" for information on the tag-based
 * format.
 */

#include <stdio.h>
#include <sys/stat.h>
#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "snap.h"
#include "load.h"

/* XXX */
put_word(FILE* fp, int w) {}
put_byte(FILE* fp, int b) {}

static void save(FILE*, unsigned char);
static void save2(FILE*, unsigned int);
static unsigned char restore(FILE*);
static unsigned int restore2(FILE*);

/* XXX */
#define z t->z80

void snap_save(char*);

int savesnap(struct trs80* t, char* fname)
{
	FILE*	fp;
	struct regz80* reg = &t->z80.reg;

	if (fname == NULL)
		fname = "trs80.snp";

	if ((fp = fopen(fname, "wb")) == NULL) {
		strcpy(loaderr_str, "Can't write to file.");
		loaderr = LD_SYSERR;
		return 0;
	}

	fprintf(fp, "%s", SNAP_MAGIC);
	put_word(fp, SNAP_MODEL_1_3);

	put_word(fp, SNAP_Z80);
	put_byte(fp, 28);

	put_word(fp, reg->AF);
	put_word(fp, reg->BC);
	put_word(fp, reg->DE);
	put_word(fp, reg->HL);
	put_word(fp, reg->AFp);
	put_word(fp, reg->BCp);
	put_word(fp, reg->DEp);
	put_word(fp, reg->HLp);
	put_word(fp, reg->IX);
	put_word(fp, reg->IY);
	put_word(fp, reg->SP);
	put_word(fp, reg->PC);
	put_byte(fp, reg->I);
	put_byte(fp, reg->R);
	put_byte(fp, reg->IFF1);
	put_byte(fp, reg->IFF2);

	put_word(fp, SNAP_MEM16);
	put_word(fp, 32768 + 2);
	put_word(fp, 0);
	fwrite(t->z80.mem, 32768, 1, fp);

	put_word(fp, SNAP_MEM16);
	put_word(fp, 32768 + 2);
	put_word(fp, 32768);
	fwrite(t->z80.mem + 32768, 32768, 1, fp);

	put_word(fp, SNAP_EOF);

	if (fclose(fp) == EOF) {
		strcpy(loaderr_str, "Error on writing.");
		loaderr = LD_SYSERR;
		return 0;
	}

	return 1;
}

int restoresnap(struct trs80* t, char* fname)
{
	FILE*	fp;
	long	count;
	struct regz80* reg;
	struct stat sinf;
	char magic[5];

	if (!(fp = path_fopen(fname, "rb")))
		return 0;

	if (fstat(fileno(fp), &sinf) == 0 && sinf.st_size == 65562) {
		reg = &t->z80.reg;
		reg->AF = restore2(fp);  reg->BC = restore2(fp);
		reg->DE = restore2(fp);  reg->HL = restore2(fp);
		reg->AFp = restore2(fp); reg->BCp = restore2(fp);
		reg->DEp = restore2(fp); reg->HLp = restore2(fp);
		reg->IX = restore2(fp);  reg->IY = restore2(fp);
		reg->SP = restore2(fp);
		reg->PC = restore2(fp);
		reg->I = restore(fp);
		reg->R = restore(fp);

		if (loaderr) {
			fclose(fp);
			return 0;
		}

		if (fread(t->z80.mem, 32768, 2, fp) != 2) {
			ueof();
			fclose(fp);
			return 0;
		}
		fclose(fp);
		return 1;
	}

	fread(magic, 4, 1, fp);
	magic[4] = '\0';
	if (strcmp(magic, SNAP_MAGIC)) {
		loaderr = 1;
		strcpy(loaderr_str, "Not a snapshot file.");
		fclose(fp);
		return 0;
	}

	return 1;
}

void save(FILE* fp, unsigned char c)
{
	fputc(c, fp);
}

void save2(FILE* fp, unsigned int x)
{
	fputc(x & 255, fp);
	fputc((x >> 8) & 255, fp);
}

unsigned char restore(FILE* fp)
{
	int ch;

	if ((ch = fgetc(fp)) == EOF) {
		ueof();
		return 0;
	}

	return ch;
}

unsigned int restore2(FILE* fp)
{
	unsigned char low;

	low = restore(fp);
	return low | (restore(fp) << 8);
}
