/*
 * z80.c -- procedural interface to the z80 simulator
 *
 * There are a few key goals here.  First, this interface should be
 * implementation independent.  For instance, a C version (rather than this
 * 80386 assembler one) would look the same to the outside code.
 * Second, it should be powerful enough that the DOS and UNIX versions
 * (and others!) can use the same interface.  Third, it is well enough
 * abstracted so that multiple Z-80s could run within the same progrm.
 * However, there's no guarantee that switching between virtual z80's
 * will not be expensive.
 * Fourth, its general form should map well onto other microprocessor
 * simulators.
 *
 * Things are a little dicey in that we have an assembler module which
 * does most of the emulation but almost none of the support routines.
 * So there's a mini-interface to the assembler part which is enhanced
 * to provide the fully functional procedural interface.
 */

#include <stdio.h>
#include <alloc.h>
#include <dos.h>

#include "mtype.h"
#include "z80.h"

/*
 * Interface to the assembler part.
 */
extern unsigned int AF;
extern unsigned int BC;
extern unsigned int DE;
extern unsigned int HL;
extern unsigned int AFp;
extern unsigned int BCp;
extern unsigned int DEp;
extern unsigned int HLp;
extern unsigned int SP;
extern unsigned int PC;
extern unsigned int IX;
extern unsigned int IY;

extern unsigned char I;
extern unsigned char R;
extern unsigned char IM;
extern unsigned char IFF1;
extern unsigned char IFF2;

extern char int_line;
extern char nmi_line;

extern char far* mem_low;
extern long cycles;

static int rawinit = 0;
extern void rawz80init(void);
extern int rawz80(void);
extern void rawz80zap(void);

static struct z80* curz80;

/* called back from the raw z80 simulator */
int inhook(int port)
{
	return (*curz80->port_read[port])(port);
}

void outhook(int port, int val)
{
	(*curz80->port_write[port])(port, val);
}


/*
 * Routines internal to this module.
 */
static mbyte mem_null(mword addr, mbyte val, mword* memloc)
{
	return val;
}

static void portwrite_null(mbyte port, mbyte val)
{
}

static mbyte portread_null(mbyte port)
{
	return 255;
}

/*
 * Routines for the world to use.
 */

int init_z80(struct z80* z)
{
	int i;

	if (!rawinit) {
		rawz80init();
		rawinit = 1;
	}
	z->reg.AF = z->reg.BC = z->reg.DE = z->reg.HL = z->reg.SP = z->reg.PC =
		z->reg.IX = z->reg.IY =	z->reg.AFp = z->reg.BCp = z->reg.DEp =
		z->reg.HLp = 0;
	z->reg.I = z->reg.R = z->reg.IM = 0; z->reg.IFF1 = z->reg.IFF2 = 0;

	z->cycles = 0;

	for (i = 0; i < 256; i++) {
		z->port_read[i] = portread_null;
		z->port_write[i] = portwrite_null;
		z->mem_read[i] = 0;
		z->mem_write[i] = 0;
	}

	if ((z->mem = farmalloc(65536 + 16)) == NULL)
		return 0;

	/* move mem_low up to a segment boundary */
	z->mem = MK_FP(FP_SEG(z->mem) + 1, 0);

	/* I'm pretty sure I have to do this in 2 parts */
	memset(z->mem, 0x76, 32768);
	memset(z->mem + 32768, 0x76, 32768);
	memset(z->mem + 15360, ' ', 1024);

	return 1;
}

/* Not sure if I want the loadup hint to stay around or not... */
int run_z80(struct z80* z, int loadup, int singlestep)
{
	int r;
	long cyclimit;

	cyclimit = 10000; /* this can't be bigger than 32767 (see z80_386.asm) */
	if (singlestep)
		cyclimit = 0;

	if (loadup) {
		AF  = z->reg.AF;  BC  = z->reg.BC;  DE  = z->reg.DE;  HL  = z->reg.HL;
		AFp = z->reg.AFp; BCp = z->reg.BCp; DEp = z->reg.DEp; HLp = z->reg.HLp;
		IX  = z->reg.IX;  IY  = z->reg.IY;
		PC  = z->reg.PC;  SP  = z->reg.SP;
		I = z->reg.I; R = z->reg.R; IM = z->reg.IM;
		IFF1 = z->reg.IFF1; IFF2 = z->reg.IFF2;
		/* Might change memory trapping substructures, too */
	}

	mem_low = z->mem;
	curz80 = z;

	for (;;) {
		cycles = cyclimit;
		r = rawz80();
		z->cycles += cyclimit - cycles;
		if (r == PROG_STOP)
			break;
		else if (r == 1) {
			r = ILL_INST;
			break;
		}
		else if (r == 2) {
			r = HALT;
			break;
		}
		else if (r == 3) {
			r = TIME_OVER;
			/* should always return, but there is no speed control yet */
			/* and it does always return, now */
			/*if (singlestep)*/
				break;
		}
	}
	z->reg.AF  = AF;  z->reg.BC  = BC;  z->reg.DE  = DE;  z->reg.HL  = HL;
	z->reg.AFp = AFp; z->reg.BCp = BCp; z->reg.DEp = DEp; z->reg.HLp = HLp;
	z->reg.IX  = IX;  z->reg.IY  = IY;
	z->reg.PC  = PC;  z->reg.SP  = SP;
	z->reg.I = I; z->reg.R = R; z->reg.IM = IM;
	z->reg.IFF1 = IFF1; z->reg.IFF2 = IFF2;

	return r;
}

/* This should really take a struct z80* parameter, but we'll assume
 * the outside program keeps track of which Z-80 is currently running.
 */
void stop_z80()
{
	rawz80zap();
}

void wportcall_z80(struct z80* z, mbyte port, portwrite callback)
{
	z->port_write[port] = callback;
}

void rportcall_z80(struct z80* z, mbyte port, portread callback)
{
	z->port_read[port] = callback;
}
